#!/bin/bash
########################################################################
# build_tensorflow.sh
#
# Build tensorflow and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_09_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}
thisdir=$(cd "${BASH_SOURCE%/*}" && pwd -P)

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: $*\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: $prog <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       $prog -h|--help [<product-dir>]

  Build $package and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking $prog.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking $prog.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "$@"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="$arg"; break
  done
  if [ -z "$product_dir" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> $product_dir\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "$product_dir/.upsfiles" ] && [ -w "$product_dir" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "$product_dir\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "$product_dir" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: $product_dir\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="$product_dir/ssibuildshims/$ssibuildshims_version"
  [ -f "$ssibuildshims_topdir/bin/ssi_functions" ] && \
    source "$ssibuildshims_topdir/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "$product_dir" && \
    setup ssibuildshims $ssibuildshims_version -z "$product_dir" || \
    ssi_die "unable to find ssibuildshims $ssibuildshims_version in
$product_dir and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "$product_dir" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  $trace verify_in_filesystem "$@" -d "$pkgdir/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "$maketar" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    $trace is_built "$@" && \
      $trace make_basic_distribution "$@" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# State variables for functions.
env_vars=()
ups_externals=()

# Add an entry to the externals file. Indicate environment variables by
# enclosing in square brackets.
add_external() {
  ups_externals+=("$1")
  shift
  local arg args=()
  for arg in "$@"; do
    [[ "$arg" =~ ^\[([^\]]+)\]$ ]] && arg="${BASH_REMATCH[1]}"
    args+=("$arg")
  done
  {
    printf "${ups_externals[@]: -1}:"
    join "/" "${args[@]}"
  } >> "$TF_UPS_EXTERNALS"
}

# Copy headers from downloaded dependencies.
copy_headers() {
  local header_file
  local header_dir
  for header_path in $(find "$1" -type f -path "$2/*/*.h" -printf "%P\n" 2>/dev/null); do
    header_dir="$incdir_full/${header_path%/*}"
    mkdir -p "$header_dir" && \
      cp -pdv "$header_path" "$header_dir/" || break;
  done
}

# Print full $BAZELRC.
print_bazelrc() {
  [ -s "$BAZELRC" ] && \
    ssi_print -i -- -1 -F '  [BAZELRC] ' -- - < "$BAZELRC" ||
      printf "<empty>\n"
}

# Print environment settings from $BAZELRC.
printenv_bazel() {
  local this_{action,cfg,comment,env}= IFS=$'\t'
  sed -Ene \
      's&^[[:space:]]*(([^[:space:]:]+)(:[^[:space:]]+)?)[[:space:]]+--action_env([[:space:]]+|=)([^=]+(=".*")?)[[:space:]]*(#.*)?$&\2\t\5\t\3\t\7&p' \
      "$BAZELRC" 2>/dev/null | while read this_action this_env this_cfg this_comment; do
    printf "  [TF_BAZEL_ENV:$this_action]${this_cfg:+\t[$this_cfg]}\t$this_env${this_comment:+ $this_comment}\n"
  done
}

# Print relevant variables in the BASH environment.
printenv_tf() {
  local var
  for var in ${env_vars[*]:+"${env_vars[@]}"}; do
    eval printf '"  [TF_ENV] '"$var=\$$var"'\n"'
  done
}

# Set a configuration-related environment variable, with optional copy
# to $BAZELRC.
setenv() {
  local action=build cfg= copy_to_bazelrc= OPT OPTIND OPTARG
  while getopts :a:bc: OPT; do
    case $OPT in
      a) action="$OPTARG";;
      b) copy_to_bazelrc=1;;
      c) cfg="$OPTARG";;
      *) ssi_fatal "usage: ${0##*/} [-b [-a <action>] [-c <cfg>]]" || return
    esac
  done
  shift $(( OPTIND - 1 ))
  OPTIND=1
  local name="${1%%=*}"
  env_vars+=("$name")
  if [ "$name" != "$1" ]; then
    local val="${1#$name}="
    export "$*" && \
      { ! (( copy_to_bazelrc )) || \
          to_bazelrc -a "$action${cfg:+:$cfg}" -e "$name=\"$val\""; }
  else
    shift
    { [ -z "$*" ] || export "$name=$*"; } && \
      { ! (( copy_to_bazelrc )) || \
          to_bazelrc -a "$action${cfg:+:$cfg}" -e "$name"${*:+="$*"}; }
  fi
}

# Write arbitrary info to $BAZELRC, one incoming word per output line.
to_bazelrc() {
  local action=build cfg= action_env= OPT OPTIND OPTARG
  while getopts :a:c:e OPT; do
    case $OPT in
      a) action="${OPTARG%%:*}"
         [ "$OPTARG" = "$action" ] || cfg="${OPTARG#*:}";;
      c) cfg="$OPTARG";;
      e) action_env=1;;
      *) ssi_fatal "usage: ${0##*/} [-a <action>[:<config>]] [-c <config>] [-e] [--] ..." || return
    esac
  done
  shift $(( OPTIND - 1 ))
  OPTIND=1
  local IFS="$IFS" OIFS="$IFS"
  if [ -z "$*" ]; then
    local msg
    IFS=$'\n'; read -rd '' -a msg < <( cat ); IFS="$OIFS"
    set -- "${msg[@]}"
  fi
  for arg in "$@"; do
    if [[ "$arg" =~ ^(([a-z][a-z_-]*)?(:[^[:space:]]+)?[[:space:]]+)?(.*)$ ]]; then
      # Deconstruct to components:
      #
      #   [action[:<config>]] [<args>...] [#<comment>]
      if [ -z "${BASH_REMATCH[4]}" ] || [ "${BASH_REMATCH[4]:0:1}" = "#" ]; then
        # Comment only.
        echo "$arg"
      else
        local this_action="${BASH_REMATCH[2]:-$action}" \
              this_component="${BASH_REMATCH[3]:-${cfg:+:$cfg}}" \
              remainder="${BASH_REMATCH[4]}" \
              comment=
        if [[ "$remainder" =~ (([[:space:]]+)|[^\\])(#.*)$ ]]; then
          comment="${BASH_REMATCH[2]}${BASH_REMATCH[3]}"
          remainder="${remainder%$comment}"
        fi
        if (( action_env )); then
          [[ "$remainder" =~ ^([^=]+)(=(.*))?$ ]] && \
            remainder="--action_env=${BASH_REMATCH[1]}${BASH_REMATCH[2]:+=${BASH_REMATCH[3]:+\"${BASH_REMATCH[3]}\"}}" || \
              ssi_fatal "unable to parse \"$remainder\" as a valid action_env setting" || return
        fi
        echo "$this_action$this_component $remainder$comment"
      fi
    else
      ssi_fatal "unable to parse \"$arg\" as a valid setting for $BAZELRC" || return
    fi
  done >> "$BAZELRC"
}

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. $thisdir/version_info.sh || \
  ssi_die "unable to load version info from $thisdir/version_info"
####################################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "$@"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,c7} \
  --pqual={p383b,p392} \
  --parse --usage-file=- -- "$@" <<EOF
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################
(( $? == 0 )) || exit

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
$trace define_basics IN_SOURCE NO_TYPE TRACK_PROGRESS || exit # *C*
##################


####################################
# Build-only dependencies. *C*

# Define versions of build-time-only products to be set up.
pybind11_default_version=v2_6_2
giflib_default_version=v5_2_1
grpc_default_version=v1_35_0
libpng_default_version=v1_6_37
libjpeg_turbo_default_version=v2_1_0
# protobuf_default_version=v3_12_3a

# Build tools.
$trace setup_cmake || exit
$trace setup_prod bazelisk -- -q+ || exit
#$trace setup_cython || exit
#$trace setup_prod numpy --latest -- -q+$(join ":+" $cqual $pqual) || exit
$trace setup_prod grpc -- -q+$cqual || exit
# Not a build tool, but header only, with no symbols or header
# inclusions exposed.
$trace setup_prod pybind11 -- -q+ || exit


# setup packages
$trace setup_prod libjpeg_turbo
$trace setup_prod giflib
$trace setup_prod libpng
##$trace setup_prod zlib

# Print active products.
echo
ups active
echo


####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
$trace fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

# Possible short-circuit if we're already built and declared.
$trace maybe_make_distribution --silent

ssi_info "building $package for $OS-$plat-$qualdir (flavor $pkgflvr)"

# Ensure directories: pkgdir, sourcedir, blddir.
$trace init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

$trace cd "$sourcedir" \
  || ssi_die "unable to cd to source directory $sourcedir"

patches=(
  ups.patch
  $pkgtag-{build,code}.patch
)

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
$trace prepare_source "${ssi_args[@]}" -P "$patchdir" \
  "$tardir/$pkgtarfile" "${patches[@]}" || exit

##################
# Tweak all inclusions of headers in packages we're getting from elsewhere:
#pkgs=(eigen protobuf) # FIXME: tensorrt / triton?
#pkgs_re="$(join '|' "${pkgs[@]}")"
#prefixes=(tensorflow/core third_party)
#prefixes=(third_party)
#prefixes_re="$(join '|' "${prefixes[@]}")"

#find . -name '*.[hHcCm]' -o -name '*.[it]cc' -o -name '*.[chit]xx' \
#  -o -name '.[chit]pp' 2>/dev/null | xargs sed -Ei'~' -e \
#  's&^([[:space:]]*#[[:space:]]*include[[:space:]]+["<])'"$prefixes_re"'/([^">]+'"$pkgs_re"'/[">].*)$&\1\2&'
##################

# Bail early if DEBUG_PATCHES is set.
[ -n "$DEBUG_PATCHES" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

# Top level build directory.
build_root="$blddir/$pkgtarname"

##################
# Python configuration and required build packages.
case $pqual in
  p383b|p392) $trace python_vars || exit;;
  *) ssi_die -I "accepted unknown pqual=$pqual"
esac

# Install some python packages needed by pip package build.
python_build_libdir="$build_root/lib/$PYTHON_LIBDIR/site_packages"
python_build_extra=(keras-{applications,preprocessing})

if (( ${#python_build_extra[@]} )); then
  $trace mkdir -p "$python_build_libdir" && \
    $trace pip install -t "$python_build_libdir" "${python_build_extra[@]}" || \
      ssi_die "unable to unstall python packages required for build (${python_build_extra[*]})"
fi
##################

# Option variables.
declare -a \
        cc_opt_flags cxx_opt_flags \
        bazel_startup_options

##################
# The principal locations for our specific configuration.
BAZELRC="$build_root/.bazelrc.user"
TF_UPS_EXTERNALS="$build_root/ups_externals.txt"
$trace rm -f "$BAZELRC" "$TF_UPS_EXTERNALS"

##################
# Write our list of external packages for use by the tensorflow build.
$trace add_external com_google_protobuf "[$PROTOBUF_FQ_DIR]" && \
  $trace add_external libjpeg_turbo "[$LIBJPEG_TURBO_DIR]" && \
  $trace add_external gif "[$GIFLIB_DIR]" && \
  $trace add_external png "[$LIBPNG_DIR]" && \
  $trace add_external eigen_archive "[$EIGEN_DIR]" && \
  $trace add_external com_github_grpc_grpc "[$GRPC_FQ_DIR]" && \
  $trace add_external org_sqlite "[$SQLITE_FQ_DIR]" && \
  $trace add_external zlib "[$ZLIB_DIR]" && \
  $trace add_external pybind11 "[$PYBIND11_DIR]" || \
    ssi_die "failed writing external package details to $TF_UPS_EXTERNALS"


#for pkg in curl; do
#  $trace add_external $pkg "[/usr]" || break;
#done || ssi_die "failed writing system package details to $TF_UPS_EXTERNALS"

##################
# Configure parallelism.
ncores=$(ncores)

# Bazel start-up (server) options.
bazel_startup_options+=(--batch --output_user_root ../build)

# Bazel build options.
$trace to_bazelrc -a build <<EOF || exit
--compilation_mode=opt
--distinct_host_configuration=false
--jobs=$ncores # Parallelism
--python_path="$python"
--subcommands=pretty_print # Bazel will print subcommands
--verbose_failures # Verbose explanation of failures
--explain="$build_root/build-explain.log" # Explain all steps

# Select configs defined in .bazelrc
--config=noaws
--config=nogcp
--config=nohdfs
--config=nonccl

# Definitions
--define=with_xla_support=false
--define=PREFIX="$pkgdir"
--define=INCLUDEDIR="$pkgdir/include"
--define=LIBDIR="$pkgdir/lib"
--define=PROTOBUF_INCLUDE_PATH="$PROTOBUF_DIR/include"

# Environment
EOF

##################
# Specify environment for specific build steps
# As variables are no longer needed in the environment outside the
# build, add their values here by appending "=<val>" to the name below.
bazelrc_env_vars=(
  LD_LIBRARY_PATH
  PROTOBUF_FQ_DIR
  PYBIND11_DIR
  PYTHONPATH
  PYTHON_BIN_PATH
  PYTHON_LIB_PATH
  SQLITE_FQ_DIR
  SWIG_FQ_DIR
  TF_SYSTEM_LIBS
  TF_UPS_EXTERNALS="$TF_UPS_EXTERNALS"
)

for ev in "${bazelrc_env_vars[@]}"; do
  $trace to_bazelrc -e "$ev" || exit
done
##################

##################
# Qualifier-specific compilation options.
case $cqual in
  e19|e20) cc=gcc; cxx=g++; cstd=11; cxxstd=17;;
  e21) cc=gcc; cxx=g++; cstd=11; cxxstd=20
       cxx_opt_flags+=(-fno-char8_t);;
  c7) cc=clang; cxx=clang++; cstd=11; cxxstd=17
      cxx_opt_flags+=(-D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1);;
  *) ssi_die -I "accepted unknown cqual=$cqual"
esac

##################
# Compiler-specific compilation options.
case $cc in
  gcc) cxx_opt_flags+=(-Wno-deprecated-copy -frecord-gcc-switches);;
  clang) cxx_opt_flags+=(-grecord-gcc-switches -stdlib=libc++);;
  *) ssi_die "unrecognized compiler $cc"
esac

# Options common to C/C++.
common_flags=(
  -fno-math-errno
  -Wno-deprecated-declarations
  -Wno-parentheses
  -Wno-sign-compare
  -msse3
)

# All C flags.
cc_opt_flags+=(${cstd:+-std=c$cstd} "${common_flags[@]}")

# All C++ flags.
cxx_opt_flags+=(
  -Wno-redundant-move
  ${cxxstd:+-std=c++$cxxstd}
  "${common_flags[@]}"
)

# adding compilation options to $BAZELRC
to_bazelrc <<EOF
# Compiler selection
--action_env=HOST_C_COMPILER="$(which $cc)"
--action_env=HOST_CXX_COMPILER="$(which $cxx)"
# C options
$(join $'\n' ${cc_opt_flags[*]:+"${cc_opt_flags[@]/#/:opt --copt=}"} \
       ${cc_opt_flags[*]:+"${cc_opt_flags[@]/#/:opt --host_copt=}"})
# C++ options
$(join $'\n' ${cxx_opt_flags[*]:+"${cxx_opt_flags[@]/#/:opt --cxxopt=}"} \
       ${cxx_opt_flags[*]:+"${cxx_opt_flags[@]/#/:opt --host_cxxopt=}"})
EOF
##################

##################
# Other build-related environment settings.

case ${cqual} in
  e*) 
    $trace setenv GCC_HOST_COMPILER_PATH="$(which gcc)"
    $trace setenv CC="$(which gcc)"
    $trace setenv CXX="$(which g++)"
    ;;
  c*)
    $trace setenv GCC_HOST_COMPILER_PATH="$(which clang)"
    $trace setenv  CC="$(which clang)"
    $trace setenv CXX="$(which clang++)"
    ;;
  *)  ssi_die -I "accepted unknown cqual=${cqual}"
esac

# Set the cache location for bazelisk and bazel.
$trace setenv XDG_CACHE_HOME "$build_root/cache"

# Environment variables used by only by deprecated configure.py.
$trace setenv TF_NEED_JEMALLOC 0
$trace setenv TF_NEED_HDFS 0
$trace setenv TF_NEED_GCP 0
$trace setenv TF_ENABLE_XLA 0
$trace setenv TF_NEED_OPENCL 0
$trace setenv TF_NEED_CUDA 0
$trace setenv TF_NEED_VERBS 0
$trace setenv TF_NEED_MKL 0
$trace setenv TF_NEED_MPI 0
$trace setenv TF_NEED_S3 0
$trace setenv TF_NEED_GDR 0
$trace setenv TF_NEED_OPENCL_SYCL 0
$trace setenv TF_SET_ANDROID_WORKSPACE false
$trace setenv TF_NEED_KAFKA 0
$trace setenv TF_NEED_AWS 0
$trace setenv TF_DOWNLOAD_CLANG 0
$trace setenv TF_NEED_IGNITE 0
$trace setenv TF_NEED_ROCM 0
$trace setenv TF_NEED_TENSORRT 0
$trace setenv TEST_TMPDIR "$build_root/tmp-test"

# These definitions should be in $BAZELRC, but currently need also to be
# set in the environment to satisfy configure.py.
$trace setenv PYTHON_BIN_PATH="$python"
$trace setenv PYTHON_LIB_PATH="$PYTHON_LIB"
$trace setenv TF_SYSTEM_LIBS "$(join ',' "${ups_externals[@]}")"
$trace setenv PYTHONPATH="$python_build_libdir:$PYTHONPATH"

##################
# Info on tensorflow-specific environment variables.
ssi_info "Tensorflow configuration environment:"
ssi_info_do printenv_tf | ssi_info -i -
ssi_info_do printf "\n"

##################
# Conditional compilation options.
#(( DEBUG_CONFIG )) || (( DEBUG_BUILD )) && compile_options+=(VERBOSE=1)
####################################

####################################
# Configure, build, and install.
#
# Use $blddir/$pkgtarname for in-source builds, $blddir otherwise.

$trace cd "$build_root" || \
  ssi_die "unable to cd to build directory $build_root"

##################
# Configure.

ssi_info "bazel user environment"
ssi_info_do printenv_bazel | ssi_info -i -
ssi_info_do printf "\n"

ssi_vinfo "full contents of $BAZELRC"
ssi_vinfo_do print_bazelrc | ssi_vinfo -i -
ssi_vinfo_do printf "\n"

# Invoke configure to get correct test settings, piping <ret> to any
# interactive questions to get defaults where necessary.
${trace} ./configure < <( yes '' ) || \
  ssi_die "unable to configure $package for build"

printf $'\n'

BAZELRC=.tf_configure.bazelrc
ssi_info "bazel environment (via ./configure_py)"
ssi_info_do printenv_bazel | ssi_info -i -
ssi_info_do printf "\n"

ssi_info "Full contents of $BAZELRC (after configuration):"
ssi_info_do print_bazelrc | ssi_info -i -
ssi_info_do printf "\n"

##################

##################
# Build. *C*

# build tensorflow library targets

bazel_targets=(
  //tensorflow:tensorflow
  //tensorflow:tensorflow_cc
  //tensorflow/tools/graph_transforms:transform_graph
  //tensorflow/compiler/tf2xla:tf2xla
  //tensorflow/compiler/xla:cpu_function_runtime
  //tensorflow/compiler/xla:executable_run_options
  //tensorflow/compiler/tf2xla:xla_compiled_cpu_function
  //tensorflow/core/profiler
  //tensorflow:install_headers
)
# Remove pip package build which requires addtiona python dependencies
#  //tensorflow/tools/pip_package:build_pip_package

for target in ${bazel_targets[*]:+"${bazel_targets[@]}"}; do
  $trace bazel ${bazel_startup_options[*]:+"${bazel_startup_options[@]}"} \
         build "$target" || \
    ssi_die "failure to build $package target $target"
done

# rebuild *.pb.{h|cc} files using the external protobuf compiler
chmod -R a+rwX $build_root/bazel-bin/tensorflow/include
for f in $(find tensorflow -name "*.proto" 2>/dev/null); do
  ${trace} protoc --cpp_out=$build_root/bazel-bin/tensorflow/include $f || \
    ssi_die "unable to rebuild protobuf files from $f"
done

# define and create empty target directories
for dir in bin include lib; do
 $trace eval ${dir:0:3}="$dir"\; ${dir:0:3}dir_full="$pkgdir/$dir"
done
$trace eval rm -rf \"\${bin,inc,lib}dir_full\" && \
  $trace eval mkdir -p \"\${bin,inc,lib}dir_full\" || \
    ssi_die - <<EOF
unable to refresh destination dirs $(join ", " $bindir $incdir $libdir) in $pkgdir
EOF

# copy targets
srcdir="$build_root/bazel-bin/tensorflow"

$trace cp -pdv "$srcdir"/{libtensorflow*.so*,compiler/tf2xla/lib*.so*,compiler/xla/lib*.so*} \
   "$libdir_full/" || ssi_die "unable to install libraries in $libdir_full"

for l in tensorflow_cc tensorflow_framework tensorflow ; do
  # check if the actual lib exists
  [ -f "$libdir_full/"lib$l.so.$pkgdotver ] || \
    ssi_die "missing expected library $libdir_full/lib$l.so.$pkgdotver"

  # link from majorversion to realversion, default lib to majorversion
  ln -sfv lib$l.so.$pkgdotver "$libdir_full/"lib$l.so.${pkgdotver%%.*} && \
    ln -sfv lib$l.so.${pkgdotver%%.*} $libdir_full/lib$l.so || \
      ssi_die "unable to create links for library $libdir_full/lib$l.so.$pkgdotver"
done

#cp -p $srcdir/compiler/aot/tfcompile $bindir
for name in tensorflow absl re2 third_party ; do
  cp -r -p "$srcdir/include/$name" "$incdir_full" || \
    ssi_die "unable to install headers for $name in $incdir_full"
done

for subdir in tensorflow/{compiler,core/profiler/{internal,lib}}; do
  copy_headers "$build_root" "$subdir" || \
    ssi_die "unable to install headers from $build_root/$subdir"
done

# build pip whl file for distributing tensorflow python package 
#$trace bazel-bin/tensorflow/tools/pip_package/build_pip_package $pkgdir || \
#  ssi_die "unable to build wheel in $pkgdir"

ssi_info "finished building $package $pkgver"

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
$trace declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
$trace is_built && $trace ssi_build_completed && \
  ssi_info "$package is installed at $pkgdir" && \
  $trace maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
